"""Performance Module"""
__docformat__ = "google"

import warnings

import pandas as pd

from financetoolkit.helpers import calculate_growth
from financetoolkit.performance import performance_model
from financetoolkit.performance.helpers import (
    handle_errors,
    handle_fama_and_french_data,
    handle_return_data_periods,
    handle_risk_free_data_periods,
)
from financetoolkit.risk.risk_model import get_ui

try:
    from tqdm import tqdm

    ENABLE_TQDM = True
except ImportError:
    ENABLE_TQDM = False

# pylint: disable=too-many-instance-attributes,too-few-public-methods,too-many-lines,too-many-locals


class Performance:
    """
    The Performance module is meant to calculate important performance metrics such
    as Sharpe Ratio, Sortino Ratio, Treynor Ratio, Information Ratio, Jensen’s
    Alpha, Beta, Capital Asset Pricing Model, R-Squared and more.
    """

    def __init__(
        self,
        tickers: str | list[str],
        historical_data: dict[str, pd.DataFrame],
        risk_free_rate_data: pd.DataFrame,
        quarterly: bool | None = None,
        rounding: int | None = 4,
        start_date: str | None = None,
        end_date: str | None = None,
        intraday_period: str | None = None,
        progress_bar: bool = True,
    ):
        """
        Initializes the Performance Controller Class.

        Args:
            tickers (str | list[str]): The tickers to use for the calculations.
            historical_data (dict[str, pd.DataFrame]): The historical data to use for the calculations.
            risk_free_rate_data (pd.DataFrame): The risk free rate data to use for the calculations.
            quarterly (bool | None, optional): Whether to use quarterly data. Defaults to None.
            rounding (int | None, optional): The number of decimals to round the results to. Defaults to 4.
            start_date (str | None, optional): The start date to use for the calculations. Defaults to None.
            end_date (str | None, optional): The end date to use for the calculations. Defaults to None.

        As an example:

        ```python
        from financetoolkit import Toolkit

        toolkit = Toolkit(["AAPL", "TSLA"], api_key="FINANCIAL_MODELING_PREP_KEY")

        toolkit.performance.get_capital_asset_pricing_model(period='quarterly')
        ```

        Which returns:

        | Date   |    AAPL |    TSLA |
        |:-------|--------:|--------:|
        | 2022Q3 | -0.0684 | -0.1047 |
        | 2022Q4 |  0.0857 |  0.0828 |
        | 2023Q1 |  0.075  |  0.1121 |
        | 2023Q2 |  0.0922 |  0.1342 |
        | 2023Q3 |  0.0052 | -0.0482 |
        """
        self._tickers = tickers
        self._benchmark_name = "Benchmark"
        self._quarterly: bool | None = quarterly
        self._rounding: int | None = rounding
        self._start_date: str | None = start_date
        self._end_date: str | None = end_date
        self._progress_bar: bool = progress_bar

        # Historical Data
        self._intraday_historical_data = historical_data["intraday"]
        self._daily_historical_data = historical_data["daily"]
        self._weekly_historical_data = historical_data["weekly"]
        self._monthly_historical_data = historical_data["monthly"]
        self._quarterly_historical_data = historical_data["quarterly"]
        self._yearly_historical_data = historical_data["yearly"]

        # Risk Free Data
        self._daily_risk_free_rate_data = risk_free_rate_data["daily"]
        self._weekly_risk_free_rate_data = risk_free_rate_data["weekly"]
        self._monthly_risk_free_rate_data = risk_free_rate_data["monthly"]
        self._quarterly_risk_free_rate_data = risk_free_rate_data["quarterly"]
        self._yearly_risk_free_rate_data = risk_free_rate_data["yearly"]

        # Fama and French
        self._fama_and_french_dataset: pd.DataFrame = pd.DataFrame()
        self._fama_and_french_model: pd.DataFrame = pd.DataFrame()
        self._fama_and_french_residuals: pd.DataFrame = pd.DataFrame()
        self._factor_asset_correlations: pd.DataFrame = pd.DataFrame()
        self._factor_correlations: pd.DataFrame = pd.DataFrame()

        # Within Period Calculations
        if not self._intraday_historical_data.empty:
            # Remove NaN values from the intraday historical data
            self._daily_within_historical_data = (
                self._intraday_historical_data.groupby(pd.Grouper(freq="D"))
                .apply(lambda x: x)
                .dropna(how="all", axis=0)
            )

            if intraday_period in ["15min", "30min", "1hour"]:
                self._intraday_within_historical_data = (
                    self._intraday_historical_data.groupby(pd.Grouper(freq="D")).apply(
                        lambda x: x
                    )
                )

                # The intraday historical data should be equal to the daily historical data
                self._intraday_historical_data = self._daily_historical_data
            else:
                self._intraday_within_historical_data = (
                    self._intraday_historical_data.groupby(pd.Grouper(freq="H")).apply(
                        lambda x: x
                    )
                )

                # The intraday historical data is grouped by hour and the last value is taken
                # and any NaN values (which occur because the market is closed) are dropped
                self._intraday_historical_data = (
                    self._intraday_historical_data.groupby(pd.Grouper(freq="H"))
                    .last()
                    .dropna()
                )

        self._weekly_within_historical_data = self._daily_historical_data.groupby(
            pd.Grouper(freq="W")
        ).apply(lambda x: x)
        self._monthly_within_historical_data = self._daily_historical_data.groupby(
            pd.Grouper(freq="M")
        ).apply(lambda x: x)
        self._quarterly_within_historical_data = self._daily_historical_data.groupby(
            pd.Grouper(freq="Q")
        ).apply(lambda x: x)
        self._yearly_within_historical_data = self._daily_historical_data.groupby(
            pd.Grouper(freq="Y")
        ).apply(lambda x: x)

        # Risk Free Rate of Intraday Historical Data is set to be equal to the last value of the daily risk free rate
        self._intraday_risk_free_rate_data = pd.Series(
            self._daily_risk_free_rate_data.iloc[-1],
            index=self._intraday_historical_data.index,
        )

    @handle_errors
    def get_beta(
        self,
        period: str | None = None,
        rolling: int | None = None,
        rounding: int | None = None,
        growth: bool = False,
        lag: int | list[int] = 1,
    ):
        """
        Calculate the Beta, a measurement that assess the systematic risk of a stock or investment.

        Beta is a financial metric used to assess the systematic risk of a stock or investment in
        relation to the overall market. It provides valuable insights into how a particular asset's
        returns tend to move in response to fluctuations in the broader market. A stock's Beta is
        calculated by analyzing its historical price movements and their correlation with the
        movements of a market index, typically the benchmark index like the S&P 500.

        The formula is as follows:

            - Beta = Covariance of Asset Returns and Benchmark Returns / Variance of Benchmark Returns

        For a given period, for example monthly, this translates into the following:

            - Beta = Monthly Covariance of Asset Returns and Benchmark Returns
            / Monthly Variance of Benchmark Returns

        See definition: https://en.wikipedia.org/wiki/Beta_(finance)

        Args:
            period (str, optional): The period to use for the calculation. Defaults to None which
            results in basing it off the quarterly parameter as defined in the class instance.
            rolling (int, optional): The rolling period to use for the calculation. If you select
            period = 'monthly' and set rolling to 12 you obtain the rolling 12-month Sharpe Ratio.
            rounding (int, optional): The number of decimals to round the results to. Defaults to 4.
            growth (bool, optional): Whether to calculate the growth of the ratios. Defaults to False.
            lag (int | str, optional): The lag to use for the growth calculation. Defaults to 1.

        Returns:
            pd.DataFrame: Beta values.

        Notes:
        - Daily Beta is not an option as the standard deviation for 1 day is close to zero. Therefore, it does
        not give any useful insights.
        - The method retrieves historical data and calculates the Beta for each asset in the Toolkit instance.
        - The risk-free rate is often represented by the return of a risk-free investment, such as a Treasury bond.
        - If `growth` is set to True, the method calculates the growth of the ratio values using the specified `lag`.

        As an example:

        ```python
        from financetoolkit import Toolkit

        toolkit = Toolkit(["AAPL", "AMZN"], api_key="FINANCIAL_MODELING_PREP_KEY")

        toolkit.performance.get_beta()
        ```
        """
        period = period if period else "quarterly" if self._quarterly else "yearly"

        historical_data = handle_return_data_periods(
            self, period, within_period=not rolling
        )
        returns = historical_data.loc[:, "Return"][self._tickers]
        benchmark_returns = historical_data.loc[:, "Return"][self._benchmark_name]

        if rolling:
            beta = performance_model.get_rolling_beta(
                returns, benchmark_returns, rolling
            )
        else:
            beta = performance_model.get_beta(returns, benchmark_returns)

        beta = beta.round(rounding if rounding else self._rounding).loc[
            self._start_date : self._end_date
        ]

        beta = beta.dropna(how="all", axis=0)

        if growth:
            return calculate_growth(
                beta,
                lag=lag,
                rounding=rounding if rounding else self._rounding,
                axis="index",
            )

        return beta

    @handle_errors
    def get_capital_asset_pricing_model(
        self,
        period: str | None = None,
        show_full_results: bool = False,
        rounding: int | None = None,
        growth: bool = False,
        lag: int | list[int] = 1,
    ):
        """
        CAPM, or the Capital Asset Pricing Model, is a financial model used to estimate the expected return
        on an investment, such as a stock or portfolio of stocks. It provides a framework for evaluating the
        risk and return trade-off of an asset or portfolio in relation to the overall market. CAPM is based
        on the following key components:

            - Risk-Free Rate (Rf): This is the theoretical return an investor could earn from an investment
            with no risk of financial loss. It is typically based on the yield of a government bond.
            - Market Risk Premium (Rm - Rf): This represents the additional return that investors expect
            to earn for taking on the risk of investing in the overall market as opposed to a risk-free asset.
            It is calculated as the difference between the expected return of the market (Rm) and the risk-free
            rate (Rf).
            - Beta (β): Beta is a measure of an asset's or portfolio's sensitivity to market movements. It
            quantifies how much an asset's returns are expected to move in relation to changes in the
            overall market. A beta of 1 indicates that the asset moves in line with the market, while a
            beta greater than 1 suggests higher volatility, and a beta less than 1 indicates lower volatility.

        The Capital Asset Pricing Model (CAPM) is a widely used financial model that helps in
        determining the expected return of an asset or portfolio based on its systematic risk and
        the prevailing risk-free rate in the market. CAPM provides insights into how an asset or
        investment should be priced in order to offer an appropriate rate of return, given its
        level of risk compared to the overall market.

        The formula is as follows:

        - Capital Asset Pricing Model = Risk Free Rate + Beta * (Benchmark Returns — Risk Free Rate)

        See definition: https://en.wikipedia.org/wiki/Capital_asset_pricing_model

        Args:
            period (str, optional): The period to use for the calculation. Defaults to None which
            results in basing it off the quarterly parameter as defined in the class instance.
            show_full_results (bool, optional): Whether to show the full results. Defaults to False.
            rounding (int, optional): The number of decimals to round the results to. Defaults to 4.
            growth (bool, optional): Whether to calculate the growth of the ratios. Defaults to False.
            lag (int | str, optional): The lag to use for the growth calculation. Defaults to 1.

        Returns:
            pd.DataFrame: CAPM values.

        Notes:
        - Daily CAPM is not an option as the standard deviation for 1 day is close to zero. Therefore, it does
        not give any useful insights.
        - The method retrieves historical data and calculates the CAPM for each asset in the Toolkit instance.
        - The risk-free rate is often represented by the return of a risk-free investment, such as a Treasury bond.
        - If `growth` is set to True, the method calculates the growth of the ratio values using the specified `lag`.

        As an example:

        ```python
        from financetoolkit import Toolkit

        toolkit = Toolkit(["AAPL", "TSLA"], api_key="FINANCIAL_MODELING_PREP_KEY")

        toolkit.performance.get_capital_asset_pricing_model()
        ```
        """
        period = period if period else "quarterly" if self._quarterly else "yearly"

        historical_data = handle_return_data_periods(self, period, within_period=True)
        returns = historical_data.loc[:, "Return"][self._tickers]
        benchmark_returns = historical_data.loc[:, "Return"][self._benchmark_name]

        beta = performance_model.get_beta(returns, benchmark_returns)

        risk_free_rate = handle_risk_free_data_periods(self, period)
        benchmark_returns = handle_return_data_periods(
            self, period, within_period=False
        ).loc[:, "Return"][self._benchmark_name]

        capm = performance_model.get_capital_asset_pricing_model(
            risk_free_rate, beta, benchmark_returns
        )

        capm = capm.round(rounding if rounding else self._rounding).loc[
            self._start_date : self._end_date
        ]

        capm = capm.dropna(how="all", axis=0)

        if show_full_results:
            full_results = pd.concat(
                [risk_free_rate, beta, benchmark_returns, capm], axis=1
            ).loc[self._start_date : self._end_date]
            full_results.columns = (
                ["Risk Free Rate"]
                + [f"Beta {ticker}" for ticker in self._tickers]
                + ["Benchmark Returns"]
                + [f"CAPM {ticker}" for ticker in self._tickers]
            )

            full_results = full_results.round(rounding if rounding else self._rounding)

            return full_results

        if growth:
            return calculate_growth(
                capm,
                lag=lag,
                rounding=rounding if rounding else self._rounding,
                axis="index",
            )

        return capm

    def get_factor_asset_correlations(
        self,
        period: str | None = None,
        factors_to_calculate: list[str] | None = None,
        rounding: int | None = None,
    ):
        """
        Calculates factor exposures for each asset.

        The major difference between the Fama and French Model here is that the correlation
        is taken as opposed to a Linear Regression in which the R-squared or Slope can be used to
        understand the exposure to each factor.

        For assessing the exposure or influence of a stock to external factors, it's often preferable
        to use R-squared (R²) or Beta because it explicitly measures how well the factors explain the stock's
        returns. A higher R² indicates that the stock's returns are more closely related to the factors,
        and thus, the factors have a greater influence on the stock's performance.

        However, since the results are closely related and tend to point into the same direction it could
        be fine to use correlations as well depending on the level of accuracy required.

        Args:
            period (str, optional): The period to use for the calculation. Defaults to None which
            results in basing it off the quarterly parameter as defined in the class instance.
            factors_to_calculate (list of str, optional): List of factors to calculate scores and residuals for.
                Defaults to ["Mkt-RF", "SMB", "HML", "RMW", "CMA"].
            rounding (int, optional): The number of decimals to round the results to. Defaults to 4.

        Returns:
            pd.DataFrame: Factor Asset Correlations.

        As an example:

        ```python
        from financetoolkit import Toolkit

        toolkit = Toolkit(["AAPL", "TSLA"], api_key="FINANCIAL_MODELING_PREP_KEY")

        toolkit.performance.get_factor_asset_correlations()
        ```
        """

        factors_to_calculate = (
            factors_to_calculate
            if factors_to_calculate
            else ["Mkt-RF", "SMB", "HML", "RMW", "CMA"]
        )

        for factor in factors_to_calculate:
            if factor not in ["Mkt-RF", "SMB", "HML", "RMW", "CMA"]:
                raise ValueError(
                    f"Factor {factor} is not a valid factor. Please select from: Mkt-RF, SMB, HML, RMW, CMA."
                )

        period = period if period else "quarterly" if self._quarterly else "yearly"

        historical_data_within = handle_return_data_periods(
            self, period, within_period=True
        )
        returns = historical_data_within.loc[:, "Return"][self._tickers]

        if self._fama_and_french_dataset.empty:
            self._fama_and_french_dataset = (
                performance_model.obtain_fama_and_french_dataset()
            )

        fama_and_french_period = handle_fama_and_french_data(
            self._fama_and_french_dataset, period
        )

        merged_df = fama_and_french_period.merge(
            returns, left_index=True, right_index=True
        )

        factor_correlations: dict = {}

        ticker_list_iterator = (
            tqdm(self._tickers, desc="Calculating Factor Asset Correlations")
            if ENABLE_TQDM & self._progress_bar
            else self._tickers
        )

        for ticker in ticker_list_iterator:
            factor_correlations[ticker] = {}
            for dataset_period in merged_df.index.get_level_values(0):
                factor_data = merged_df.loc[dataset_period][factors_to_calculate]
                excess_returns = (
                    merged_df.loc[dataset_period][ticker]
                    - merged_df.loc[dataset_period]["RF"]
                )

                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    factor_correlations[ticker][
                        dataset_period
                    ] = performance_model.get_factor_asset_correlations(
                        factors=factor_data, excess_return=excess_returns
                    )

        factor_asset_correlations = pd.DataFrame.from_dict(
            {
                (ticker, dataset_period): value
                for ticker, factor_scores_ticker in factor_correlations.items()
                for dataset_period, value in factor_scores_ticker.items()
            },
        )

        factor_order = factor_asset_correlations.index

        factor_asset_correlations = (
            factor_asset_correlations.stack(level=1)
            .unstack(level=0)
            .reindex(factor_order, level=1, axis=1)
            .reindex(self._tickers, level=0, axis=1)
        )

        self._factor_asset_correlations = factor_asset_correlations.round(
            rounding if rounding else self._rounding
        ).loc[self._start_date : self._end_date]

        return self._factor_asset_correlations

    def get_factor_correlations(
        self,
        period: str | None = None,
        factors_to_calculate: list[str] | None = None,
        exclude_risk_free: bool = True,
        rounding: int | None = None,
    ):
        """
        Calculates factor correlations between each factor. This is useful to understand how correlated
        each factor is to each other. This is based off the Fama and French 5 Factor model which includes:

            - Market Risk Premium (Mkt-RF): Represents the additional return that investors expect to earn for taking
            on the risk of investing in the overall market as opposed to a risk-free asset.
            - Size Premium (SMB): Reflects the historical excess return of small-cap stocks over large-cap stocks.
            - Value Premium (HML): Captures the historical excess return of value stocks over growth stocks.
            - Profitability (RMW): Measures the historical excess return of high profitability stocks over
            low profitability stocks.
            - Investment (CMA): Quantifies the historical excess return of low investment stocks over
            high investment stocks.

        Optionally, it is also possible to see the correlation between the risk-free rate and each factor.

        Args:
            period (str, optional): The period to use for the calculation. Defaults to None which
            results in basing it off the quarterly parameter as defined in the class instance.
            factors_to_calculate (list of str, optional): List of factors to calculate scores and residuals for.
                Defaults to ["Mkt-RF", "SMB", "HML", "RMW", "CMA"].
            exclude_risk_free (bool, optional): Whether to exclude the risk-free rate from the results. Defaults to True.
            rounding (int, optional): The number of decimals to round the results to. Defaults to 4.

        Returns:
            pd.DataFrame: Factor Correlations.

        As an example:

        ```python
        from financetoolkit import Toolkit

        toolkit = Toolkit(["AAPL", "TSLA"], api_key="FINANCIAL_MODELING_PREP_KEY")

        toolkit.performance.get_factor_correlations()
        ```
        """
        factors_to_calculate = (
            factors_to_calculate
            if factors_to_calculate
            else ["Mkt-RF", "SMB", "HML", "RMW", "CMA"]
        )

        for factor in factors_to_calculate:
            if factor not in ["Mkt-RF", "SMB", "HML", "RMW", "CMA"]:
                raise ValueError(
                    f"Factor {factor} is not a valid factor. Please select from: Mkt-RF, SMB, HML, RMW, CMA."
                )

        period = period if period else "quarterly" if self._quarterly else "yearly"

        factors_to_calculate = (
            factors_to_calculate + ["RF"]
            if not exclude_risk_free
            else factors_to_calculate
        )

        if self._fama_and_french_dataset.empty:
            self._fama_and_french_dataset = (
                performance_model.obtain_fama_and_french_dataset()
            )

        fama_and_french_period = handle_fama_and_french_data(
            self._fama_and_french_dataset[factors_to_calculate],
            period,
            correlation=True,
        )

        self._factor_correlations = fama_and_french_period.round(
            rounding if rounding else self._rounding
        )

        return self._factor_correlations

    def get_fama_and_french_model(
        self,
        period: str | None = None,
        method: str = "multi",
        factors_to_calculate: list[str] | None = None,
        include_daily_residuals: bool = False,
        rounding: int | None = None,
        growth: bool = False,
        lag: int | list[int] = 1,
    ):
        """
        Calculate Fama and French 5 Factor model scores and residuals for a set of financial assets.

        The Fama and French 5 Factor model is a widely used financial model that helps estimate the expected return
        of financial assets, such as stocks or portfolios, based on five key factors:

            - Market Risk Premium (Mkt-RF): Represents the additional return that investors expect to earn for taking
            on the risk of investing in the overall market as opposed to a risk-free asset.
            - Size Premium (SMB): Reflects the historical excess return of small-cap stocks over large-cap stocks.
            - Value Premium (HML): Captures the historical excess return of value stocks over growth stocks.
            - Profitability (RMW): Measures the historical excess return of high profitability stocks over
            low profitability stocks.
            - Investment (CMA): Quantifies the historical excess return of low investment stocks over
            high investment stocks.

        The model can perform both a Simple Linear Regression on each factor as well as a Multi Linear Regression
        which includes all factors. Generally, a multi linear regression is applied but if you wish to see individual
        R-squared values for each factor you can select the simple linear regression method.

        The model performs a Linear Regression on each factor and defines the regression parameters and residuals
        for each asset over time based on its exposure to these factors.

        These results can be validated by comparing them to the period returns obtained from the historical data. E.g.
        the regression formula is as follows for the Multi Linear Regression:

            - Excess Return = Intercept + Beta1 * Mkt-RF + Beta2 * SMB + Beta3 * HML + Beta4 * RMW
                + Beta5 * CMA + Residuals

        And the following for the Simple Linear Regression:

            - Excess Return = Intercept + Slope * Factor Value + Residuals

        So for a given factor, it should hold that the Excess Return equals the entire regression. Note that in this
        calculation the Excess Return refers to the Asset Return minus the Risk Free Rate as reported in the Fama and
        French dataset and will not be the same as the defined Excess Return in the historical data given that this is
        based on the Risk Free Rate defined in the initialization.

        What is relevant to look at is the influence these factors have on each stock and how much each factor explains
        the stock return. E.g. you will generally see a pretty high influence (Beta or Slope) for the Market Risk Premium
        (Mkt-RF) factor as this is the main factor that explains the stock return (as also prevalent in the CAPM).
        The other factors can fluctuate greatly between stocks depending on which stocks you look at.

        Args:
            period (str, optional): The period for the calculation (e.g., "weekly", "monthly", "quarterly", "yearly").
                Defaults to None, using class-defined quarterly or yearly period.
            method (str, optional): The regression method to use for the calculation. Defaults to 'multi'.
            factors_to_calculate (list of str, optional): List of factors to calculate scores and residuals for.
                Defaults to ["Mkt-RF", "SMB", "HML", "RMW", "CMA"].
            include_residuals (bool, optional): Whether to include residuals in the results. Defaults to False.
            rounding (int, optional): The number of decimals to round the results to. Defaults to 4.
            growth (bool, optional): Whether to calculate the growth of the ratio values. Defaults to False.
            lag (int or list of int, optional): The lag to use for the growth calculation. Defaults to 1.

        Returns:
            pd.DataFrame: Fama and French 5 Factor model scores for the specified assets.

        Notes:
        - The dataset from Fama and French is not always fully up to date. Therefore, some periods could be excluded.
        - Daily Fama and French results is not an option as it would attempt to do a linear regression on a
        single data point which will not give any meaningful results.
        - The method retrieves historical data and calculates regression parameters and residuals for each asset.
        - The risk-free rate is typically represented by the return of a risk-free investment, such as a Treasury bond.
        In this case, the Risk Free Rate from the Fama and French dataset is used.
        - If `growth` is set to True, the method calculates the growth of the ratio values using the specified `lag`.

        As an example:

        ```python
        from financetoolkit import Toolkit

        toolkit = Toolkit(["AAPL", "TSLA"], api_key="FINANCIAL_MODELING_PREP_KEY")

        # Calculate Fama and French 5 Factor model scores
        toolkit.performance.get_fama_and_french_model()
        ```
        """
        if method not in ["simple", "multi"]:
            raise ValueError(
                f"Method {method} is not a valid method. Please select from: simple or multi (default). "
                "This refers to a simple linear regression (regression of each factor on each asset return series) "
                "or a multi linear regression (regression of all factors on each asset return series) respectively"
            )

        factors_to_calculate = (
            factors_to_calculate
            if factors_to_calculate
            else ["Mkt-RF", "SMB", "HML", "RMW", "CMA"]
        )

        for factor in factors_to_calculate:
            if factor not in ["Mkt-RF", "SMB", "HML", "RMW", "CMA"]:
                raise ValueError(
                    f"Factor {factor} is not a valid factor. Please select from: Mkt-RF, SMB, HML, RMW, CMA."
                )

        period = period if period else "quarterly" if self._quarterly else "yearly"

        historical_data_within = handle_return_data_periods(
            self, period, within_period=True
        )
        returns = historical_data_within.loc[:, "Return"][self._tickers]

        historical_data = handle_return_data_periods(self, period, within_period=False)
        returns_total = historical_data.loc[:, "Return"][self._tickers]

        self._fama_and_french_dataset = (
            performance_model.obtain_fama_and_french_dataset()
        )
        fama_and_french_period = handle_fama_and_french_data(
            self._fama_and_french_dataset, period
        )

        merged_df = fama_and_french_period.merge(
            returns, left_index=True, right_index=True
        )
        factor_scores: dict = {}
        daily_residuals: dict = {}

        ticker_list_iterator = (
            tqdm(
                self._tickers,
                desc=f"Calculating {'Multi' if method == 'multi' else 'Individual'} Factor Exposures",
            )
            if ENABLE_TQDM & self._progress_bar
            else self._tickers
        )

        for ticker in ticker_list_iterator:
            factor_scores[ticker] = {}
            daily_residuals[ticker] = {}

            if method == "multi":
                for dataset_period in merged_df.index.get_level_values(0):
                    factor_data = merged_df.loc[dataset_period][factors_to_calculate]
                    excess_returns = (
                        merged_df.loc[dataset_period][ticker]
                        - merged_df.loc[dataset_period]["RF"]
                    )

                    (
                        factor_scores[ticker][dataset_period],
                        daily_residuals[ticker][dataset_period],
                    ) = performance_model.get_fama_and_french_model_multi(
                        excess_returns=excess_returns, factor_dataset=factor_data
                    )

                    fama_and_french_model = pd.DataFrame.from_dict(
                        {
                            (ticker, factor): value
                            for ticker, factor_scores_ticker in factor_scores.items()
                            for factor, value in factor_scores_ticker.items()
                        },
                        orient="index",
                    )

                fama_and_french_model = fama_and_french_model.unstack(
                    level=0, sort=False
                ).swaplevel(0, 1, axis=1)

                # Sort the DataFrame with respect to the original column order
                tickers_column_order = fama_and_french_model.columns.get_level_values(
                    0
                ).unique()
                parameters_column_order = (
                    fama_and_french_model.columns.get_level_values(1).unique()
                )

                fama_and_french_model = (
                    fama_and_french_model.sort_index(axis=1)
                    .reindex(tickers_column_order, level=0, axis=1)
                    .reindex(parameters_column_order, level=1, axis=1)
                )

            elif method == "simple":
                for factor in factors_to_calculate:
                    factor_scores[ticker][factor] = {}
                    daily_residuals[ticker][factor] = {}
                    for dataset_period in merged_df.index.get_level_values(0):
                        factor_data = merged_df.loc[dataset_period][factor]
                        excess_returns = (
                            merged_df.loc[dataset_period][ticker]
                            - merged_df.loc[dataset_period]["RF"]
                        )

                        (
                            factor_scores[ticker][factor][dataset_period],
                            daily_residuals[ticker][factor][dataset_period],
                        ) = performance_model.get_fama_and_french_model_single(
                            excess_returns=excess_returns, factor=factor_data
                        )

                        factor_scores[ticker][factor][dataset_period][
                            "Factor Value"
                        ] = factor_data.iloc[-1]

                        factor_scores[ticker][factor][dataset_period][
                            "Residuals"
                        ] = returns_total.loc[dataset_period][ticker] - (
                            factor_scores[ticker][factor][dataset_period]["Slope"]
                            * factor_data.iloc[-1]
                            + factor_scores[ticker][factor][dataset_period]["Intercept"]
                        )

                fama_and_french_model = pd.DataFrame.from_dict(
                    {
                        (period, factor, ticker): value
                        for ticker, factor_scores_ticker in factor_scores.items()
                        for factor, factor_scores_factor in factor_scores_ticker.items()
                        for period, value in factor_scores_factor.items()
                    },
                    orient="index",
                )

                # Sort the DataFrame with respect to the original column order
                parameters_column_order = fama_and_french_model.columns.unique()
                factor_column_order = fama_and_french_model.index.get_level_values(
                    1
                ).unique()
                ticker_column_order = fama_and_french_model.index.get_level_values(
                    2
                ).unique()

                fama_and_french_model = fama_and_french_model.stack(sort=False).unstack(
                    level=[2, 1, 3]
                )

                fama_and_french_model = (
                    fama_and_french_model.sort_index(axis=1)
                    .reindex(parameters_column_order, level=2, axis=1)
                    .reindex(factor_column_order, level=1, axis=1)
                    .reindex(ticker_column_order, level=0, axis=1)
                )

        self._fama_and_french_model = fama_and_french_model.round(
            rounding if rounding else self._rounding
        ).loc[self._start_date : self._end_date]

        if include_daily_residuals:
            if method == "multi":
                daily_residuals_df = pd.DataFrame.from_dict(
                    {
                        (ticker, factor): value
                        for ticker, residuals_ticker in daily_residuals.items()
                        for factor, value in residuals_ticker.items()
                    },
                    orient="index",
                )

                daily_residuals_df = (
                    daily_residuals_df.stack()
                    .unstack(level=0)
                    .sort_index(axis=1, sort_remaining=False)
                )
                daily_residuals_df = daily_residuals_df.reset_index(level=0, drop=True)
            else:
                daily_residuals_df = pd.DataFrame.from_dict(
                    {
                        (period, factor, ticker): value
                        for ticker, residuals_ticker in daily_residuals.items()
                        for factor, residuals_factor in residuals_ticker.items()
                        for period, value in residuals_factor.items()
                    },
                    orient="index",
                )

                daily_residuals_df = (
                    daily_residuals_df.unstack(level=[1, 2])
                    .stack(level=0)
                    .sort_index(axis=1, sort_remaining=False)
                )

                daily_residuals_df = daily_residuals_df.reset_index(level=0, drop=True)

            self._fama_and_french_residuals = daily_residuals_df.round(
                rounding if rounding else self._rounding
            ).loc[self._start_date : self._end_date]

            return self._fama_and_french_model, self._fama_and_french_residuals

        if growth:
            return calculate_growth(
                self._fama_and_french_model,
                lag=lag,
                rounding=rounding if rounding else self._rounding,
                axis="index",
            )

        return self._fama_and_french_model

    @handle_errors
    def get_alpha(
        self,
        period: str | None = None,
        show_full_results: bool = False,
        rounding: int | None = None,
        growth: bool = False,
        lag: int | list[int] = 1,
    ):
        """
        Alpha, in a general sense, represents the excess return an investment generates relative to
        a benchmark or a risk-adjusted return. It can be positive (indicating the investment
        outperformed the benchmark) or negative (indicating underperformance).

        The formula is as follows:

            - Alpha = Asset's Actual Return — Benchmark's Actual Return

        See definition: https://en.wikipedia.org/wiki/Alpha_(finance)

        Args:
            period (str, optional): The period to use for the calculation. Defaults to None which
            results in basing it off the quarterly parameter as defined in the class instance.
            rounding (int, optional): The number of decimals to round the results to. Defaults to 4.
            growth (bool, optional): Whether to calculate the growth of the ratios. Defaults to False.
            lag (int | str, optional): The lag to use for the growth calculation. Defaults to 1.

        Returns:
            pd.DataFrame: Alpha values.

        Notes:
        - The method retrieves historical data and calculates the Alpha for each asset in the Toolkit instance.
        - If `growth` is set to True, the method calculates the growth of the ratio values using the specified `lag`.

        As an example:

        ```python
        from financetoolkit import Toolkit

        toolkit = Toolkit(["AAPL", "TSLA"], api_key="FINANCIAL_MODELING_PREP_KEY")

        toolkit.performance.get_alpha()
        ```
        """
        period = period if period else "quarterly" if self._quarterly else "yearly"

        historical_data = handle_return_data_periods(self, period, within_period=False)
        returns = historical_data.loc[:, "Return"][self._tickers]
        benchmark_returns = historical_data.loc[:, "Return"][self._benchmark_name]

        alpha = performance_model.get_alpha(returns, benchmark_returns)

        alpha = alpha.round(rounding if rounding else self._rounding).loc[
            self._start_date : self._end_date
        ]

        alpha = alpha.dropna(how="all", axis=0)

        if show_full_results:
            full_results = pd.concat([returns, benchmark_returns, alpha], axis=1).loc[
                self._start_date : self._end_date
            ]
            full_results.columns = (
                [f"Actual Return {ticker}" for ticker in self._tickers]
                + ["Benchmark Returns"]
                + [f"Alpha {ticker}" for ticker in self._tickers]
            )

            return full_results

        if growth:
            return calculate_growth(
                alpha,
                lag=lag,
                rounding=rounding if rounding else self._rounding,
                axis="index",
            )

        return alpha

    def get_jensens_alpha(
        self,
        period: str | None = None,
        rounding: int | None = None,
        growth: bool = False,
        lag: int | list[int] = 1,
    ):
        """
        Calculate Jensen's Alpha, a measure of an asset's performance relative to its expected return
        based on the Capital Asset Pricing Model (CAPM).

        Jensen's Alpha is used to assess whether an investment has outperformed or underperformed
        its expected return given its systematic risk, as represented by the asset's Beta.

        The formula is as follows:

            - Jensen's Alpha = Asset's Actual Return — [Risk-Free Rate + Beta * (Benchmark Return — Risk-Free Rate)]

        See definition: https://en.wikipedia.org/wiki/Jensen%27s_alpha

        Args:
            period (str, optional): The period to use for the calculation. Defaults to None which
            results in basing it off the quarterly parameter as defined in the class instance.
            rounding (int, optional): The number of decimals to round the results to. Defaults to 4.
            growth (bool, optional): Whether to calculate the growth of the ratios. Defaults to False.
            lag (int | str, optional): The lag to use for the growth calculation. Defaults to 1.

        Returns:
            pd.DataFrame: Jensen's Alpha values.

        Notes:
        - Daily Jensen's Alpha is not an option as the standard deviation for 1 day is close to zero. Therefore, it does
        not give any useful insights.
        - The method retrieves historical data and calculates the CAPM for each asset in the Toolkit instance.
        - The risk-free rate is often represented by the return of a risk-free investment, such as a Treasury bond.
        - If `growth` is set to True, the method calculates the growth of the ratio values using the specified `lag`.

        As an example:

        ```python
        from financetoolkit import Toolkit

        toolkit = Toolkit(["AAPL", "TSLA"], api_key="FINANCIAL_MODELING_PREP_KEY")

        toolkit.performance.get_jensens_alpha()
        ```
        """
        period = period if period else "quarterly" if self._quarterly else "yearly"

        historical_data = handle_return_data_periods(self, period, within_period=True)
        returns = historical_data.loc[:, "Return"][self._tickers]
        benchmark_returns = historical_data.loc[:, "Return"][self._benchmark_name]

        beta = performance_model.get_beta(returns, benchmark_returns)

        historical_period_data = handle_return_data_periods(
            self, period, within_period=False
        )
        period_returns = historical_period_data.loc[:, "Return"][self._tickers]

        risk_free_rate = handle_risk_free_data_periods(self, period)
        benchmark_returns = handle_return_data_periods(
            self, period, within_period=False
        ).loc[:, "Return"][self._benchmark_name]

        jensens_alpha = performance_model.get_jensens_alpha(
            period_returns, risk_free_rate, beta, benchmark_returns
        )

        jensens_alpha = jensens_alpha.round(
            rounding if rounding else self._rounding
        ).loc[self._start_date : self._end_date]

        jensens_alpha = jensens_alpha.dropna(how="all", axis=0)

        if growth:
            return calculate_growth(
                jensens_alpha,
                lag=lag,
                rounding=rounding if rounding else self._rounding,
                axis="index",
            )

        return jensens_alpha

    @handle_errors
    def get_treynor_ratio(
        self,
        period: str | None = None,
        rounding: int | None = None,
        growth: bool = False,
        lag: int | list[int] = 1,
    ):
        """
        The Treynor Ratio, also known as Treynor's Measure or the Reward-to-Variability Ratio, is a
        financial metric used to assess the risk-adjusted performance of an investment portfolio or
        asset. It measures the excess return generated by the portfolio per unit of systematic or market
        risk, often represented by Beta. The Treynor Ratio is a valuable tool for evaluating the
        performance of investments in relation to their market risk exposure.

        The formula is as follows:

        - - Treynor Ratio = (Portfolio's Return — Risk-Free Rate) / Portfolio Beta

        See definition: https://en.wikipedia.org/wiki/Treynor_ratio

        Args:
            period (str, optional): The period to use for the calculation. Defaults to None which
            results in basing it off the quarterly parameter as defined in the class instance.
            rounding (int, optional): The number of decimals to round the results to. Defaults to 4.
            growth (bool, optional): Whether to calculate the growth of the ratios. Defaults to False.
            lag (int | str, optional): The lag to use for the growth calculation. Defaults to 1.

        Returns:
            pd.DataFrame: Treynor Ratio values.

        Notes:
        - Daily Treynor Ratio is not an option as the standard deviation for 1 day is close to zero. Therefore,
        it does not give any useful insights.
        - The method retrieves historical data and calculates the TR for each asset in the Toolkit instance.
        - The risk-free rate is often represented by the return of a risk-free investment, such as a Treasury bond.
        - If `growth` is set to True, the method calculates the growth of the ratio values using the specified `lag`.

        As an example:

        ```python
        from financetoolkit import Toolkit

        toolkit = Toolkit(["AAPL", "TSLA"], api_key="FINANCIAL_MODELING_PREP_KEY")

        toolkit.performance.get_treynor_ratio()
        ```
        """
        period = period if period else "quarterly" if self._quarterly else "yearly"

        historical_data = handle_return_data_periods(self, period, within_period=True)
        returns = historical_data.loc[:, "Return"][self._tickers]
        benchmark_returns = historical_data.loc[:, "Return"][self._benchmark_name]

        beta = performance_model.get_beta(returns, benchmark_returns)

        historical_period_data = handle_return_data_periods(
            self, period, within_period=False
        )
        period_returns = historical_period_data.loc[:, "Return"][self._tickers]
        risk_free_rate = handle_risk_free_data_periods(self, period)

        treynor_ratio = performance_model.get_treynor_ratio(
            period_returns, risk_free_rate, beta
        )

        treynor_ratio = treynor_ratio.round(
            rounding if rounding else self._rounding
        ).loc[self._start_date : self._end_date]

        treynor_ratio = treynor_ratio.dropna(how="all", axis=0)

        if growth:
            return calculate_growth(
                treynor_ratio,
                lag=lag,
                rounding=rounding if rounding else self._rounding,
                axis="index",
            )

        return treynor_ratio

    @handle_errors
    def get_sharpe_ratio(
        self,
        period: str | None = None,
        rolling: int | None = None,
        rounding: int | None = None,
        growth: bool = False,
        lag: int | list[int] = 1,
    ):
        """
        Calculate the Sharpe ratio, a measure of risk-adjusted return that evaluates the excess return
        of an investment portfolio or asset per unit of risk taken.

        The Sharpe ratio is calculated as the difference between the expected return of the asset or portfolio
        and the risk-free rate of return, divided by the standard deviation of the asset or portfolio's excess return.
        It quantifies the amount of return generated for each unit of risk assumed, providing insights into the
        investment's performance relative to the risk taken.

        The formula is as follows:

            - Sharpe Ratio = Excess Return / Excess Standard Deviation

        For a given period, for example monthly, this translates into the following:

            - Sharpe Ratio = Average Monthly Excess Return / Standard Deviation of Monthly Excess Returns

        For a rolling period, this translates into the following:

            - Sharpe Ratio = Average Rolling Excess Return / Standard Deviation of Rolling Excess Returns

        Note that this is explicitly already subtracts the Risk Free Rate.

        See definition: https://en.wikipedia.org/wiki/Sharpe_ratio

        Args:
            period (str, optional): The period to use for the calculation. Defaults to None which
            results in basing it off the quarterly parameter as defined in the class instance.
            rolling (int, optional): The rolling period to use for the calculation. If you select
            period = 'monthly' and set rolling to 12 you obtain the rolling 12-month Sharpe Ratio.
            rounding (int, optional): The number of decimals to round the results to. Defaults to 4.
            growth (bool, optional): Whether to calculate the growth of the ratios. Defaults to False.
            lag (int | str, optional): The lag to use for the growth calculation. Defaults to 1.

        Returns:
            pd.DataFrame: Sharpe ratio values.

        Notes:
        - Daily Sharpe Ratio is not an option as the standard deviation for 1 day is close to zero. Therefore, it does
        not give any useful insights.
        - The method retrieves historical data and calculates the Sharpe ratio for each asset in the Toolkit instance.
        - The risk-free rate is often represented by the return of a risk-free investment, such as a Treasury bond.
        - If `growth` is set to True, the method calculates the growth of the ratio values using the specified `lag`.

        As an example:

        ```python
        from financetoolkit import Toolkit

        toolkit = Toolkit(["AAPL", "TSLA"], api_key="FINANCIAL_MODELING_PREP_KEY")

        toolkit.performance.get_sharpe_ratio()
        ```
        """
        period = period if period else "quarterly" if self._quarterly else "yearly"

        historical_data = handle_return_data_periods(
            self, period, within_period=not rolling
        )
        excess_return = historical_data.loc[:, "Excess Return"][self._tickers]

        if rolling:
            sharpe_ratio = performance_model.get_rolling_sharpe_ratio(
                excess_return, rolling
            )
        else:
            sharpe_ratio = performance_model.get_sharpe_ratio(excess_return)

        sharpe_ratio = sharpe_ratio.round(rounding if rounding else self._rounding).loc[
            self._start_date : self._end_date
        ]

        sharpe_ratio = sharpe_ratio.dropna(how="all", axis=0)

        if growth:
            return calculate_growth(
                sharpe_ratio,
                lag=lag,
                rounding=rounding if rounding else self._rounding,
                axis="index",
            )

        return sharpe_ratio

    @handle_errors
    def get_sortino_ratio(
        self,
        period: str | None = None,
        rounding: int | None = None,
        growth: bool = False,
        lag: int | list[int] = 1,
    ):
        """

        The Sortino Ratio is a financial metric used to assess the risk-adjusted performance of an
        investment portfolio or asset by considering only the downside risk. It measures the excess
        return generated by the portfolio per unit of downside risk, specifically, the standard
        deviation of negative returns. The Sortino Ratio is particularly useful for investors who
        are primarily concerned with minimizing the downside risk of their investments.

        The formula is as follows:

            - Sortino Ratio = Excess Return / Excess Downside Risk

        For a given period, for example monthly, this translates into the following:

            - Sortino Ratio = Average Monthly Excess Return / Average Monthly Excess Downside Risk

        For a rolling period, this translates into the following:

            - Sortino Ratio = Average Rolling Excess Return / Rolling Downside Risk

        Note that this is explicitly already subtracts the Risk Free Rate.

        See definition: https://en.wikipedia.org/wiki/Sortino_ratio

        Args:
            period (str, optional): The period to use for the calculation. Defaults to None which
            results in basing it off the quarterly parameter as defined in the class instance.
            rounding (int, optional): The number of decimals to round the results to. Defaults to 4.
            growth (bool, optional): Whether to calculate the growth of the ratios. Defaults to False.
            lag (int | str, optional): The lag to use for the growth calculation. Defaults to 1.

        Returns:
            pd.DataFrame: Sortino ratio values.

        Notes:
        - Daily Sortino Ratio is not an option as the standard deviation for 1 day is close to zero. Therefore, it does
        not give any useful insights.
        - The method retrieves historical data and calculates the Sortino ratio for each asset in the Toolkit instance.
        - The risk-free rate is often represented by the return of a risk-free investment, such as a Treasury bond.
        - If `growth` is set to True, the method calculates the growth of the ratio values using the specified `lag`.

        As an example:

        ```python
        from financetoolkit import Toolkit

        toolkit = Toolkit(["AAPL", "TSLA"], api_key="FINANCIAL_MODELING_PREP_KEY")

        toolkit.performance.get_sortino_ratio()
        ```
        """
        period = period if period else "quarterly" if self._quarterly else "yearly"

        historical_data = handle_return_data_periods(self, period, within_period=True)
        excess_return = historical_data.loc[:, "Excess Return"][self._tickers]

        sortino_ratio = performance_model.get_sortino_ratio(excess_return)

        sortino_ratio = sortino_ratio.round(
            rounding if rounding else self._rounding
        ).loc[self._start_date : self._end_date]

        sortino_ratio = sortino_ratio.dropna(how="all", axis=0)

        if growth:
            return calculate_growth(
                sortino_ratio,
                lag=lag,
                rounding=rounding if rounding else self._rounding,
                axis="index",
            )

        return sortino_ratio

    @handle_errors
    def get_ulcer_performance_index(
        self,
        period: str | None = None,
        rolling: int = 14,
        rounding: int | None = None,
        growth: bool = False,
        lag: int | list[int] = 1,
    ):
        """
        Calculate the Ulcer Performance Index (UPI), alternatively called Martin ratio, a measure of risk-adjusted
        return that evaluates the excess return of an investment portfolio or asset per unit of risk taken.

        It can be used to compare volatilities in different stocks or show stocks go into Ulcer territory.
        Similar to the Sharpe Ratio, a higher UPI is better than a lower one (since investors prefer more return
        for less risk).

        Args:
            period (str, optional): The period to use for the calculation. Defaults to None which
            results in basing it off the quarterly parameter as defined in the class instance.
            rolling (int): The rolling period to use to calculate the Ulcer Index. Defaults to 14.
            rounding (int, optional): The number of decimals to round the results to. Defaults to 4.
            growth (bool, optional): Whether to calculate the growth of the ratios. Defaults to False.
            lag (int | str, optional): The lag to use for the growth calculation. Defaults to 1.

        Returns:
            pd.DataFrame: Ulcer Performance Index values.

        Notes:
        - The method retrieves historical data and calculates the UPI for each asset in the Toolkit instance.
        - The risk-free rate is often represented by the return of a risk-free investment, such as a Treasury bond.
        - If `growth` is set to True, the method calculates the growth of the ratio values using the specified `lag`.

        As an example:

        ```python
        from financetoolkit import Toolkit

        toolkit = Toolkit(["AAPL", "TSLA"], api_key="FINANCIAL_MODELING_PREP_KEY")

        toolkit.performance.get_ulcer_performance_index()
        ```
        """

        period = period if period else "quarterly" if self._quarterly else "yearly"
        return_column = "Return" if period == "intraday" else "Excess Return"

        historical_data = handle_return_data_periods(self, period, True)
        returns = historical_data.loc[:, "Return"][self._tickers]
        historical_data_within_period = handle_return_data_periods(self, period, False)
        excess_return = historical_data_within_period.loc[:, return_column][
            self._tickers
        ]

        ulcer_index = get_ui(returns, rolling)

        ulcer_performance_index = performance_model.get_ulcer_performance_index(
            excess_return, ulcer_index
        )
        ulcer_performance_index = ulcer_performance_index.round(
            rounding if rounding else self._rounding
        ).loc[self._start_date : self._end_date]

        ulcer_performance_index = ulcer_performance_index.dropna(how="all", axis=0)

        if growth:
            return calculate_growth(
                ulcer_performance_index,
                lag=lag,
                rounding=rounding if rounding else self._rounding,
                axis="index",
            )

        return ulcer_performance_index

    @handle_errors
    def get_m2_ratio(
        self,
        period: str | None = None,
        rounding: int | None = None,
        growth: bool = False,
        lag: int | list[int] = 1,
    ):
        """
        The M2 Ratio, also known as the Modigliani-Modigliani Measure, is a financial
        metric used to evaluate the risk-adjusted performance of an investment portfolio
        or strategy. It assesses the excess return generated by the portfolio relative
        to a risk-free investment, taking into account the portfolio's volatility or risk.
        The M2 Ratio helps investors and portfolio managers determine whether the portfolio
        is delivering returns that justify its level of risk.

        The formula is as follows:

            - M2 Ratio = (Portfolio's Return — Risk-Free Rate) / Portfolio Standard Deviation

        See definition: https://en.wikipedia.org/wiki/Modigliani_risk-adjusted_performance

        Args:
            period (str, optional): The period to use for the calculation. Defaults to None which
            results in basing it off the quarterly parameter as defined in the class instance.
            rounding (int, optional): The number of decimals to round the results to. Defaults to 4.
            growth (bool, optional): Whether to calculate the growth of the ratios. Defaults to False.
            lag (int | str, optional): The lag to use for the growth calculation. Defaults to 1.

        Returns:
            pd.DataFrame: M2 ratio values.

        Notes:
        - Daily M2 is not an option as the standard deviation for 1 day is close to zero. Therefore, it does
        not give any useful insights.
        - The method retrieves historical data and calculates the M2 for each asset in the Toolkit instance.
        - The risk-free rate is often represented by the return of a risk-free investment, such as a Treasury bond.
        - If `growth` is set to True, the method calculates the growth of the ratio values using the specified `lag`.

        As an example:

        ```python
        from financetoolkit import Toolkit

        toolkit = Toolkit(["AAPL", "TSLA"], api_key="FINANCIAL_MODELING_PREP_KEY")

        toolkit.performance.get_m2_ratio()
        ```
        """
        period = period if period else "quarterly" if self._quarterly else "yearly"

        historical_period_data = handle_return_data_periods(
            self, period, within_period=False
        )
        period_returns = historical_period_data.loc[:, "Return"][self._tickers]
        period_standard_deviation = historical_period_data.loc[:, "Volatility"][
            self._tickers
        ]
        risk_free_rate = handle_risk_free_data_periods(self, period)

        m2_ratio = performance_model.get_m2_ratio(
            period_returns, risk_free_rate, period_standard_deviation
        )

        m2_ratio = m2_ratio.round(rounding if rounding else self._rounding).loc[
            self._start_date : self._end_date
        ]

        m2_ratio = m2_ratio.dropna(how="all", axis=0)

        if growth:
            return calculate_growth(
                m2_ratio,
                lag=lag,
                rounding=rounding if rounding else self._rounding,
                axis="index",
            )

        return m2_ratio

    @handle_errors
    def get_tracking_error(
        self,
        period: str | None = None,
        rounding: int | None = None,
        growth: bool = False,
        lag: int | list[int] = 1,
    ):
        """
        Tracking Error is a financial metric that quantifies the volatility or dispersion of the
        difference between the returns of an investment portfolio or asset and the returns of a
        benchmark index. It measures how closely the portfolio tracks its benchmark and provides
        insights into the consistency of the portfolio's performance relative to the benchmark.
        A higher Tracking Error indicates greater divergence from the benchmark, while a lower
        Tracking Error suggests that the portfolio closely follows the benchmark.

        The formula is as follows:

            - Tracking Error (TE) = Standard Deviation of (Portfolio Returns — Benchmark Returns)

        See definition: https://en.wikipedia.org/wiki/Tracking_error

        Args:
            period (str, optional): The period to use for the calculation. Defaults to None which
            results in basing it off the quarterly parameter as defined in the class instance.
            rounding (int, optional): The number of decimals to round the results to. Defaults to 4.
            growth (bool, optional): Whether to calculate the growth of the ratios. Defaults to False.
            lag (int | str, optional): The lag to use for the growth calculation. Defaults to 1.

        Returns:
            pd.DataFrame: Tracking error values.

        Notes:
        - Daily Tracking Error is not an option as the standard deviation for 1 day is close to zero. Therefore, it does
        not give any useful insights.
        - The method retrieves historical data and calculates the TE for each asset in the Toolkit instance.
        - The risk-free rate is often represented by the return of a risk-free investment, such as a Treasury bond.
        - If `growth` is set to True, the method calculates the growth of the ratio values using the specified `lag`.

        As an example:

        ```python
        from financetoolkit import Toolkit

        toolkit = Toolkit(["AAPL", "TSLA"], api_key="FINANCIAL_MODELING_PREP_KEY")

        toolkit.performance.get_tracking_error()
        ```
        """
        period = period if period else "quarterly" if self._quarterly else "yearly"

        historical_data = handle_return_data_periods(self, period, within_period=True)
        returns = historical_data.loc[:, "Return"][self._tickers]
        benchmark_returns = historical_data.loc[:, "Return"][self._benchmark_name]

        tracking_error = performance_model.get_tracking_error(
            returns, benchmark_returns
        )

        tracking_error = tracking_error.round(
            rounding if rounding else self._rounding
        ).loc[self._start_date : self._end_date]

        tracking_error = tracking_error.dropna(how="all", axis=0)

        if growth:
            return calculate_growth(
                tracking_error,
                lag=lag,
                rounding=rounding if rounding else self._rounding,
                axis="index",
            )

        return tracking_error

    @handle_errors
    def get_information_ratio(
        self,
        period: str | None = None,
        rounding: int | None = None,
        growth: bool = False,
        lag: int | list[int] = 1,
    ):
        """
        The Information Ratio (IR), also known as the Information Coefficient, is a financial
        metric that assesses the risk-adjusted performance of a portfolio or investment
        strategy relative to a benchmark index. It quantifies how much excess return the
        portfolio generates for each unit of tracking error (volatility of tracking error).
        The Information Ratio is commonly used by portfolio managers, financial analysts, and
        investors to evaluate the skill of a portfolio manager in generating returns beyond
        what would be expected based on the risk taken.

        - IR > 0: A positive Information Ratio indicates that the portfolio has generated
        excess returns compared to the benchmark, suggesting that the portfolio manager has added value.
        - IR = 0: An Information Ratio of zero implies that the portfolio's excess return
        is in line with the benchmark, meaning the portfolio manager has not added or lost value relative
        to the benchmark.
        - IR < 0: A negative Information Ratio suggests that the portfolio has underperformed
        the benchmark, potentially indicating that the portfolio manager has detracted value.

        The formula is as follows:

            - Information Ratio (IR) = (Portfolio's Excess Return — Benchmark's Excess Return) / Tracking Error

        See definition: https://en.wikipedia.org/wiki/Information_ratio

        Args:
            period (str, optional): The period to use for the calculation. Defaults to None which
            results in basing it off the quarterly parameter as defined in the class instance.
            rounding (int, optional): The number of decimals to round the results to. Defaults to 4.
            growth (bool, optional): Whether to calculate the growth of the ratios. Defaults to False.
            lag (int | str, optional): The lag to use for the growth calculation. Defaults to 1.

        Returns:
            pd.DataFrame: Information ratio values.

        Notes:
        - Daily Information Ratio is not an option as the standard deviation for 1 day is close to zero.
        Therefore, it does not give any useful insights.
        - The method retrieves historical data and calculates the IR for each asset in the Toolkit instance.
        - The risk-free rate is often represented by the return of a risk-free investment, such as a Treasury bond.
        - If `growth` is set to True, the method calculates the growth of the ratio values using the specified `lag`.

        As an example:

        ```python
        from financetoolkit import Toolkit

        toolkit = Toolkit(["AAPL", "TSLA"], api_key="FINANCIAL_MODELING_PREP_KEY")

        toolkit.performance.get_information_ratio()
        ```
        """
        period = period if period else "quarterly" if self._quarterly else "yearly"

        historical_data = handle_return_data_periods(self, period, within_period=True)
        returns = historical_data.loc[:, "Return"][self._tickers]
        benchmark_returns = historical_data.loc[:, "Return"][self._benchmark_name]

        information_ratio = performance_model.get_information_ratio(
            returns, benchmark_returns
        )

        information_ratio = information_ratio.round(
            rounding if rounding else self._rounding
        ).loc[self._start_date : self._end_date]

        if growth:
            return calculate_growth(
                information_ratio,
                lag=lag,
                rounding=rounding if rounding else self._rounding,
                axis="index",
            )

        return information_ratio

    @handle_errors
    def get_compound_growth_rate(
        self,
        rounding: int | None = None,
    ):
        """
        This function calculates the Compound Growth Rate (CGR) for different periods: yearly, quarterly, monthly,
        weekly, and daily.

        The CGR is a measure that provides the mean growth rate of an investment over a specified period of time.
        It is a useful measure for comparing the performance of investments over different time periods or across
        different asset classes. The CGR is calculated by taking the ratio of the final value to the initial value,
        raising it to the inverse of the number of periods, and then subtracting one.

        The formula is as follows:

            - CGR = (Final Value / Initial Value) ^ (1 / Number of Periods) — 1

        Args:
            rounding (int, optional): The number of decimals to round the results to. If not provided,
            the function will use the default rounding value set in the class instance.

        Returns:
            pd.DataFrame: A DataFrame containing the CGR for each period. The DataFrame has the periods
            as the index and the CGR values as the column.

        Notes:
        - When verifying the calculation, note that rounding applies and it could be slightly off because of that
        This is mostly noticeable when looking at the Compound Daily Growth Rate. Adjust the rounding with the
        rounding parameter accordingly to get a more precise figure.

        As an example:

        ```python
        from financetoolkit import Toolkit

        toolkit = Toolkit(["AAPL", "TSLA"], api_key="FINANCIAL_MODELING_PREP_KEY")

        toolkit.performance.get_compound_growth_rate()
        ```
        """
        prices = self._yearly_historical_data.loc[:, "Adj Close"].loc[
            self._start_date : self._end_date
        ]

        cagr = performance_model.get_compound_growth_rate(prices, len(prices))

        prices = self._quarterly_historical_data.loc[:, "Adj Close"].loc[
            self._start_date : self._end_date
        ]

        cqgr = performance_model.get_compound_growth_rate(prices, len(prices))

        prices = self._monthly_historical_data.loc[:, "Adj Close"].loc[
            self._start_date : self._end_date
        ]

        cqgr = performance_model.get_compound_growth_rate(prices, len(prices))

        prices = self._weekly_historical_data.loc[:, "Adj Close"].loc[
            self._start_date : self._end_date
        ]

        cwgr = performance_model.get_compound_growth_rate(prices, len(prices))

        prices = self._daily_historical_data.loc[:, "Adj Close"].loc[
            self._start_date : self._end_date
        ]

        cdgr = performance_model.get_compound_growth_rate(prices, len(prices))

        compound_growth_rate = pd.DataFrame(
            [cagr, cqgr, cqgr, cwgr, cdgr],
            index=[
                "Compound Annual Growth Rate (CAGR)",
                "Compound Quarterly Growth Rate (CQGR)",
                "Compound Monthly Growth Rate (CMGR)",
                "Compound Weekly Growth Rate (CWGR)",
                "Compound Daily Growth Rate (CDGR)",
            ],
        )

        compound_growth_rate = compound_growth_rate.round(
            rounding if rounding else self._rounding
        )

        return compound_growth_rate



"""Growth Model"""
__docformat__ = "google"

import pandas as pd

# pylint: disable=too-many-locals


def get_present_value_of_growth_opportunities(
    weighted_average_cost_of_capital: pd.DataFrame,
    earnings_per_share: pd.DataFrame,
    close_prices: pd.DataFrame,
    calculate_daily: bool = False,
) -> pd.DataFrame:
    """
    The Present Value of Growth Opportunities (PVGO) is the net present value of all future investments a company is
    expected to make. It is calculated as the difference between the Close Price and the Earnings Per Share divided by
    the Weighted Average Cost of Capital.

    It is meant to be used as relative valuation metric and therefore doesn't necessarily have a meaning when used
    for one company.


    The formula is as follows:

        PVGO = Close Price — (Earnings Per Share / Weighted Average Cost of Capital)

    Args:
        weighted_average_cost_of_capital (pd.DataFrame): The weighted average cost of capital.
        earnings_per_share (pd.DataFrame): The earnings per share.
        close_prices (pd.DataFrame): The close prices.
        calculate_daily (bool): Whether to calculate the PVGO on a daily basis. If False, the PVGO is calculated
        based on the provided close_prices DataFrame.

    Returns:
        pd.DataFrame: The PVGO.
    """
    earnings_wacc_ratio = (earnings_per_share / weighted_average_cost_of_capital).T

    if calculate_daily:
        earnings_wacc_ratio.index = pd.PeriodIndex(earnings_wacc_ratio.index, freq="D")
        earnings_wacc_ratio = pd.DataFrame(
            earnings_wacc_ratio, index=close_prices.index
        ).ffill()

    return close_prices - earnings_wacc_ratio



"""OECD Model"""
__docformat__ = "google"

import pandas as pd

# pylint: disable=too-many-lines

BASE_URL = "https://stats.oecd.org/sdmx-json/data/DP_LIVE/"
EXTENSIONS = "/OECD?contentType=csv&detail=code&separator=comma&csv-lang=en"

CODE_TO_COUNTRY = {
    "AGO": "Angola",
    "ALB": "Albania",
    "ARE": "United Arab Emirates",
    "ARG": "Argentina",
    "ARM": "Armenia",
    "AUS": "Australia",
    "AUT": "Austria",
    "AZE": "Azerbaijan",
    "BEL": "Belgium",
    "BEN": "Benin",
    "BFA": "Burkina Faso",
    "BGD": "Bangladesh",
    "BGR": "Bulgaria",
    "BHR": "Bahrain",
    "BIH": "Bosnia and Herzegovina",
    "BLR": "Belarus",
    "BOL": "Bolivia",
    "BRA": "Brazil",
    "BRN": "Brunei",
    "BWA": "Botswana",
    "CAN": "Canada",
    "CHE": "Switzerland",
    "CHL": "Chile",
    "CHN": "China",
    "CIV": "Ivory Coast",
    "CMR": "Cameroon",
    "COD": "Democratic Republic of the Congo",
    "COG": "Republic of the Congo",
    "COL": "Colombia",
    "CRI": "Costa Rica",
    "CUB": "Cuba",
    "CYP": "Cyprus",
    "CZE": "Czech Republic",
    "DEU": "Germany",
    "DEW": "Former Federal Republic of Germany",
    "DNK": "Denmark",
    "DOM": "Dominican Republic",
    "DZA": "Algeria",
    "EA": "Euro Area",
    "EA17": "Euro Area 17",
    "EA19": "Euro Area 19",
    "ECU": "Ecuador",
    "EGY": "Egypt",
    "ERI": "Eritrea",
    "ESP": "Spain",
    "EST": "Estonia",
    "ETH": "Ethiopia",
    "EU": "European Union",
    "EU27_2020": "European Union 27",
    "EU28": "European Union 28",
    "FIN": "Finland",
    "FJI": "Fiji",
    "FRA": "France",
    "G-20": "G20",
    "G-7": "G7",
    "G7M": "G7M",
    "GAB": "Gabon",
    "GBR": "United Kingdom",
    "GEO": "Georgia",
    "GHA": "Ghana",
    "GNQ": "Equatorial Guinea",
    "GRC": "Greece",
    "GTM": "Guatemala",
    "HKG": "Hong Kong",
    "HND": "Honduras",
    "HRV": "Croatia",
    "HTI": "Haiti",
    "HUN": "Hungary",
    "IDN": "Indonesia",
    "IND": "India",
    "IRL": "Ireland",
    "IRN": "Iran",
    "IRQ": "Iraq",
    "ISL": "Iceland",
    "ISR": "Israel",
    "ITA": "Italy",
    "JAM": "Jamaica",
    "JOR": "Jordan",
    "JPN": "Japan",
    "KAZ": "Kazakhstan",
    "KEN": "Kenya",
    "KGZ": "Kyrgyzstan",
    "KHM": "Cambodia",
    "KOR": "South Korea",
    "KWT": "Kuwait",
    "LAO": "Laos",
    "LBN": "Lebanon",
    "LBY": "Libya",
    "LIE": "Liechtenstein",
    "LKA": "Sri Lanka",
    "LSO": "Lesotho",
    "LTU": "Lithuania",
    "LUX": "Luxembourg",
    "LVA": "Latvia",
    "MAR": "Morocco",
    "MDA": "Moldova",
    "MDG": "Madagascar",
    "MEX": "Mexico",
    "MKD": "Macedonia",
    "MLI": "Mali",
    "MLT": "Malta",
    "MMR": "Myanmar",
    "MNE": "Montenegro",
    "MNG": "Mongolia",
    "MOZ": "Mozambique",
    "MRT": "Mauritania",
    "MUS": "Mauritius",
    "MWI": "Malawi",
    "MYS": "Malaysia",
    "NAM": "Namibia",
    "NER": "Niger",
    "NGA": "Nigeria",
    "NIC": "Nicaragua",
    "NLD": "Netherlands",
    "NMEC": "Non-OECD Economies",
    "NOR": "Norway",
    "NPL": "Nepal",
    "NZL": "New Zealand",
    "OAVG": "OECD - Average",
    "OECD": "OECD - Total",
    "OECDE": "OECD - Europe",
    "OEU": "OECD - Europe",
    "OMN": "Oman",
    "PAK": "Pakistan",
    "PAN": "Panama",
    "PER": "Peru",
    "PHL": "Philippines",
    "PNG": "Papua New Guinea",
    "POL": "Poland",
    "PRK": "North Korea",
    "PRT": "Portugal",
    "PRY": "Paraguay",
    "PSE": "Palestine",
    "QAT": "Qatar",
    "ROU": "Romania",
    "RUS": "Russia",
    "RWA": "Rwanda",
    "SAU": "Saudi Arabia",
    "SDN": "Sudan",
    "SEN": "Senegal",
    "SGP": "Singapore",
    "SLE": "Sierra Leone",
    "SLV": "El Salvador",
    "SRB": "Serbia",
    "SVK": "Slovakia",
    "SVN": "Slovenia",
    "SWE": "Sweden",
    "SWZ": "Swaziland",
    "SYR": "Syria",
    "TCD": "Chad",
    "TGO": "Togo",
    "THA": "Thailand",
    "TJK": "Tajikistan",
    "TKM": "Turkmenistan",
    "TTO": "Trinidad and Tobago",
    "TUN": "Tunisia",
    "TUR": "Turkey",
    "TWN": "Taiwan",
    "TZA": "Tanzania",
    "UGA": "Uganda",
    "UKR": "Ukraine",
    "URY": "Uruguay",
    "USA": "United States",
    "UZB": "Uzbekistan",
    "VEN": "Venezuela",
    "VNM": "Vietnam",
    "WLD": "World",
    "YEM": "Yemen",
    "ZAF": "South Africa",
    "ZMB": "Zambia",
    "ZWE": "Zimbabwe",
}


def collect_oecd_data(oecd_data_string: str, period_code: str) -> pd.DataFrame:
    """
    Collect the data from the OECD API and return it as a DataFrame. This is
    a helper function for the other functions in this module.

    Args:
        oece_data_string (str): The string that is appended to the base URL to
            get the data from the OECD API.
        period_code (str): The period code of the data. Can be 'M' for monthly,
            'Q' for quarterly or 'Y' for yearly.

    Returns:
       pd.DataFrame: A DataFrame containing the data from the OECD API.
    """
    oecd_data = pd.read_csv(f"{BASE_URL}{oecd_data_string}{EXTENSIONS}")

    oecd_data["LOCATION"] = oecd_data["LOCATION"].replace(CODE_TO_COUNTRY)

    oecd_data = oecd_data.drop(
        columns=["INDICATOR", "SUBJECT", "MEASURE", "FREQUENCY", "Flag Codes"]
    )

    oecd_data = oecd_data.drop_duplicates(subset=["TIME", "LOCATION"])

    oecd_data = oecd_data.pivot(index="TIME", columns="LOCATION", values="Value")

    oecd_data.index.name = None
    oecd_data.columns.name = None

    oecd_data.index = pd.PeriodIndex(oecd_data.index, freq=period_code)

    # Only remove data if all of it is NaN
    oecd_data = oecd_data.dropna(axis=1, how="all")

    return oecd_data


def get_annual_gross_domestic_product(per_capita: bool = False) -> pd.DataFrame:
    """
    Get the Gross Domestic Product for a variety of countries over
    time from the OECD. The Gross Domestic Product is the total value
    of goods produced and services provided in a country during one year.

    The data is available in two forms: per capita and total. The per capita
    data is the GDP divided by the population of the country. The total data
    is the GDP in millions of dollars.

    Args:
        per_capita (bool): Whether to return the per capita data or the total data.

    Returns:
       pd.DataFrame: A DataFrame containing the Gross Domestic Product for a variety
       of countries over time.
    """
    oecd_data_string = (
        f".GDP.TOT.{'USD_CAP' if per_capita else 'MLN_USD'}.A{EXTENSIONS}"
    )

    gross_domestic_product = collect_oecd_data(oecd_data_string, "Y")

    return gross_domestic_product


def get_quarterly_gross_domestic_product(year_on_year: bool = True) -> pd.DataFrame:
    """
    Get the Quarterly Gross Domestic Product for a variety of countries over
    time from the OECD. The Gross Domestic Product is the total value
    of goods produced and services provided in a country during one year.

    The data is available in two forms: compared to the previous year's value or
    compared to the previous period. The year on year data is the GDP compared to
    the same quarter in the previous year. The quarter on quarter data is the GDP
    compared to the previous quarter.

    Args:
        per_capita (bool): Whether to return the per capita data or the total data.

    Returns:
       pd.DataFrame: A DataFrame containing the Quarterly Gross Domestic Product
       for a variety of countries over time.
    """
    oecd_data_string = f".QGDP.TOT.{'PC_CHGPY' if year_on_year else 'PC_CHGPP'}.Q"

    gross_domestic_product = collect_oecd_data(oecd_data_string, "Q")

    # Division by 100 due to both metrics are in percentages
    gross_domestic_product = gross_domestic_product / 100

    return gross_domestic_product


def get_gross_domestic_product_forecast_short_term(
    inflation_adjusted: bool = False, quarterly: bool = False
) -> pd.DataFrame:
    """
    Get the Gross Domestic Product Forecast for a variety of countries over
    time from the OECD. The Gross Domestic Product is the total value
    of goods produced and services provided in a country during one year.

    The data represents the annual growth rate which is the percentage change
    in GDP from one year to the next. When using quarterly data, the annual
    growth rate is the percentage change in GDP from the same quarter in the
    previous year.

    Args:
        per_capita (bool): Whether to return the per capita data or the total data.

    Returns:
       pd.DataFrame: A DataFrame containing the Gross Domestic Product for a variety
       of countries over time.
    """
    oecd_data_string = (
        f".{'REALGDPFORECAST' if inflation_adjusted else 'NOMGDPFORECAST'}.TOT."
        f"AGRWTH.{'Q' if quarterly else 'A'}"
    )

    gross_domestic_product = collect_oecd_data(
        oecd_data_string, "Q" if quarterly else "Y"
    )

    # Division by 100 due to both metrics are in percentages
    gross_domestic_product = gross_domestic_product / 100

    return gross_domestic_product


def get_gross_domestic_product_forecast_long_term():
    """
    Get the Gross Domestic Product Forecast for a variety of countries over
    time from the OECD. The Gross Domestic Product is the total value
    of goods produced and services provided in a country during one year.

    The data represents the millions of US dollars in GDP over a long
    period of time (> 40 years) which resembles an estimation of the
    total GDP over time.

    Returns:
       pd.DataFrame: A DataFrame containing the Gross Domestic Product for a variety
       of countries over time.
    """
    oecd_data_string = ".GDPLTFORECAST.TOT.MLN_USD.A"

    gross_domestic_product = collect_oecd_data(oecd_data_string, "Y")

    return gross_domestic_product


def get_consumer_confidence_index():
    """
    Get the Consumer Confidence Index for a variety of countries over
    time from the OECD. The Consumer Confidence Index is an indicator
    that measures the degree of optimism that consumers feel about
    the overall state of the economy and their personal financial
    situation. Data is defined in months.

    Returns:
       pd.DataFrame: A DataFrame containing the Consumer Confidence Index for a variety
        of countries over time.
    """
    oecd_data_string = ".CCI.AMPLITUD.LTRENDIDX.M"

    consumer_confidence_index = collect_oecd_data(oecd_data_string, "M")

    return consumer_confidence_index


def get_business_confidence_index():
    """
    Get the Business Confidence Index for a variety of countries over
    time from the OECD. The Business Confidence Index is an indicator
    that measures the degree of optimism that businesses feel about
    the overall state of the economy and their personal financial
    situation. Data is defined in months.

    Returns:
       pd.DataFrame: A DataFrame containing the Business Confidence Index for a variety
        of countries over time.
    """
    oecd_data_string = ".BCI.AMPLITUD.LTRENDIDX.M"

    business_confidence_index = collect_oecd_data(oecd_data_string, "M")

    return business_confidence_index


def get_composite_leading_indicator():
    """
    Get the Composite Leading Indicator for a variety of countries over
    time from the OECD. The Composite Leading Indicator is an indicator
    that tries to determine the turning points in business cycles.
    Data is defined in months.

    Returns:
       pd.DataFrame: A DataFrame containing the Composite Leading Indicator for a variety
        of countries over time.
    """
    oecd_data_string = ".CLI.AMPLITUD.LTRENDIDX.M"

    composite_leading_indicator = collect_oecd_data(oecd_data_string, "M")

    return composite_leading_indicator


def get_consumer_price_index(period: str, inflation_type: str):
    """
    Get the Consumer Price Index for a variety of countries over
    time from the OECD. The Consumer Price Index is an indicator
    that measures the average change over time in the prices paid
    by urban consumers for a market basket of consumer goods and
    services.

    Args:
        period (str): The period of the data. Can be 'monthly', 'quarterly' or 'yearly'.
        inflation_type (str): The type of inflation. Can be 'total', 'food' or 'energy'.

    Returns:
       pd.DataFrame: A DataFrame containing the Consumer Price Index for a variety
        of countries over time.
    """
    inflation_type = inflation_type.lower()
    period = period.lower()

    if inflation_type not in ["total", "total_excl", "food", "energy"]:
        raise ValueError(
            "Inflation type must be one of 'total', 'total_excl', 'food' or 'energy'"
        )
    if period not in ["monthly", "quarterly", "yearly"]:
        raise ValueError("Period must be one of 'monthly', 'quarterly' or 'yearly'")

    inflation_data = (
        "TOT"
        if inflation_type == "total"
        else "FOOD"
        if inflation_type == "food"
        else "TOT_FOODENRG"
        if inflation_type == "total_excl"
        else "ENRG"
    )
    period_data = "M" if period == "monthly" else "Q" if period == "quarterly" else "A"

    oecd_data_string = f".CPI.{inflation_data}.IDX2015.{period_data}"

    consumer_price_index = collect_oecd_data(
        oecd_data_string, "Y" if period_data == "A" else period_data
    )

    return consumer_price_index


def get_consumer_price_index_forecast(period: str):
    """
    Get the Consumer Price Index Forecast which is a growth rate
    for a variety of countries over time from the OECD. The
    Consumer Price Index is an indicator that measures the average
    change over time in the prices paid by urban consumers for a
    market basket of consumer goods and services.

    Args:
        period (str): The period of the data. Can be 'monthly', 'quarterly' or 'yearly'.
        inflation_type (str): The type of inflation. Can be 'total', 'food' or 'energy'.

    Returns:
       pd.DataFrame: A DataFrame containing the Consumer Price Index for a variety
        of countries over time.
    """
    period = period.lower()

    if period not in ["quarterly", "yearly"]:
        raise ValueError("Period must be 'quarterly' or 'yearly'")

    period_data = "M" if period == "monthly" else "Q" if period == "quarterly" else "A"

    oecd_data_string = f".CPIFORECAST.TOT.AGRWTH.{period_data}"

    consumer_price_index = collect_oecd_data(
        oecd_data_string, "Y" if period_data == "A" else period_data
    )

    # Division by 100 due to being a growth rate
    consumer_price_index = consumer_price_index / 100

    return consumer_price_index


def get_producer_price_index(period: str):
    """
    Get the Producer Price Index for a variety of countries over
    time from the OECD. The Producer Price Index is an indicator
    that measures the average change over time in the selling
    prices received by domestic producers for their output.

    Args:
        period (str): The period of the data. Can be 'monthly', 'quarterly' or 'yearly'.

    Returns:
       pd.DataFrame: A DataFrame containing the Producer Price Index for a variety
        of countries over time.
    """
    period = period.lower()

    if period not in ["monthly", "quarterly", "yearly"]:
        raise ValueError("Period must be one of 'monthly', 'quarterly' or 'yearly'")

    period_data = "M" if period == "monthly" else "Q" if period == "quarterly" else "A"

    oecd_data_string = f".PPI.TOT_MKT.IDX2015.{period_data}"

    producer_price_index = collect_oecd_data(
        oecd_data_string, "Y" if period_data == "A" else period_data
    )

    return producer_price_index


def get_house_prices(quarterly: bool | None = False, inflation_adjusted: bool = False):
    """
    Get the house prices for a variety of countries over time from the OECD.
    The house prices are the prices of residential real estate in the country.

    Args:
        quarterly (bool): Whether to return the quarterly data or the yearly data.
        inflation_adjusted (bool): Whether to return the inflation adjusted data or the
            nominal data.

    Returns:
       pd.DataFrame: A DataFrame containing the house prices for a variety
        of countries over time.
    """
    oecd_data_string = f".HOUSECOST.{'REAL' if inflation_adjusted else 'NOMINAL'}.IDX2015.{'Q' if quarterly else 'A'}"

    house_prices = collect_oecd_data(oecd_data_string, "Q" if quarterly else "Y")

    return house_prices


def get_rent_prices(
    quarterly: bool | None = False,
):
    """
    Get the rent prices for a variety of countries over time from the OECD.
    The rent prices are the prices of residential real estate in the country.

    Args:
        quarterly (bool): Whether to return the quarterly data or the yearly data.

    Returns:
       pd.DataFrame: A DataFrame containing the rent prices for a variety
        of countries over time.
    """
    oecd_data_string = f".HOUSECOST.RENT.IDX2015.{'Q' if quarterly else 'A'}"

    house_prices = collect_oecd_data(oecd_data_string, "Q" if quarterly else "Y")

    return house_prices


def get_unemployment_rate(period: str):
    """
    Get the unemployment rate for a variety of countries over time from the OECD.
    The unemployment rate is the percentage of the total labor force that is unemployed.

    Returns:
       pd.DataFrame: A DataFrame containing the unemployment rate for a variety
        of countries over time.
    """
    period = period.lower()

    if period not in ["monthly", "quarterly", "yearly"]:
        raise ValueError("Period must be one of 'monthly', 'quarterly' or 'yearly'")

    period_data = "M" if period == "monthly" else "Q" if period == "quarterly" else "A"

    oecd_data_string = f".HUR.TOT.PC_LF.{period_data}"

    unemployment_rate = collect_oecd_data(
        oecd_data_string, "Y" if period_data == "A" else period_data
    )

    # Divide by 100 to get percentage
    unemployment_rate = unemployment_rate / 100

    return unemployment_rate


def get_long_term_interest_rate(period: str, forecast: bool = False):
    """
    Get the long term interest rate for a variety of countries over time from the OECD.
    The long term interest rate is defined as the yield on government bonds with a
    maturity of 10 years.

    Args:
        period (str): The period of the data. Can be 'monthly', 'quarterly' or 'yearly'.
        forecast (bool): Whether to return the forecast data or the historical data.

    Returns:
       pd.DataFrame: A DataFrame containing the long term interest rate for a variety
        of countries over time.
    """
    period = period.lower()

    if period not in ["monthly", "quarterly", "yearly"]:
        raise ValueError("Period must be one of 'monthly', 'quarterly' or 'yearly'")

    period_data = "M" if period == "monthly" else "Q" if period == "quarterly" else "A"

    oecd_data_string = (
        f".{'LTINTFORECAST' if forecast else 'LTINT'}.TOT.PC_PA.{period_data}"
    )

    long_term_interest_rate = collect_oecd_data(
        oecd_data_string, "Y" if period_data == "A" else period_data
    )

    # Divide by 100 to get percentage
    long_term_interest_rate = long_term_interest_rate / 100

    return long_term_interest_rate


def get_short_term_interest_rate(period: str, forecast: bool = False):
    """
    Get the short term interest rate for a variety of countries over time from the OECD.
    The short term interest rate is defined as the yield on government bonds with a
    maturity of 3 months.

    Args:
        period (str): The period of the data. Can be 'monthly', 'quarterly' or 'yearly'.
        forecast (bool): Whether to return the forecast data or the historical data.

    Returns:
       pd.DataFrame: A DataFrame containing the short term interest rate for a variety
        of countries over time.
    """
    period = period.lower()

    if period not in ["monthly", "quarterly", "yearly"]:
        raise ValueError("Period must be one of 'monthly', 'quarterly' or 'yearly'")

    period_data = "M" if period == "monthly" else "Q" if period == "quarterly" else "A"

    oecd_data_string = (
        f".{'STINTFORECAST' if forecast else 'STINT'}.TOT.PC_PA.{period_data}"
    )

    short_term_interest_rate = collect_oecd_data(
        oecd_data_string, "Y" if period_data == "A" else period_data
    )

    # Divide by 100 to get percentage
    short_term_interest_rate = short_term_interest_rate / 100

    return short_term_interest_rate


def get_purchasing_power_parity():
    """
    Get the purchasing power parity for a variety of countries over time from the OECD.
    Purchasing power parity (PPP) is a measure of the price of specific goods in
    different countries and is used to compare the absolute purchasing power of
    the countries' currencies.

    Returns:
       pd.DataFrame: A DataFrame containing the purchasing power parity for a variety
        of countries over time.
    """
    oecd_data_string = ".PPP.TOT.NATUSD.A"

    purchasing_power_parity = collect_oecd_data(oecd_data_string, "Y")

    return purchasing_power_parity


def get_exchange_rates():
    """
    Get the exchange rates for a variety of countries over time from the OECD.
    The exchange rate is defined as the number of units of the national currency
    that are needed to purchase one unit of US dollars.

    Returns:
       pd.DataFrame: A DataFrame containing the exchange rates for a variety
        of countries over time.
    """
    oecd_data_string = ".EXCH.TOT.NATUSD.A"

    exchange_rates = collect_oecd_data(oecd_data_string, "Y")

    return exchange_rates


def get_share_prices(period: str):
    """
    Share price indices are calculated from the prices of common shares of
    companies traded on national or foreign stock exchanges. They are usually
    determined by the stock exchange, using the closing daily values for the
    monthly data, and normally expressed as simple arithmetic averages of
    the daily data.

    Args:
        period (str): The period of the data. Can be 'monthly', 'quarterly' or 'yearly'.

    Returns:
       pd.DataFrame: A DataFrame containing the share prices for a variety
        of countries over time.
    """
    period = period.lower()

    if period not in ["monthly", "quarterly", "yearly"]:
        raise ValueError("Period must be one of 'monthly', 'quarterly' or 'yearly'")

    period_data = "M" if period == "monthly" else "Q" if period == "quarterly" else "A"

    oecd_data_string = f".SHPRICE.TOT.IDX2015.{period_data}"

    share_prices = collect_oecd_data(
        oecd_data_string, "Y" if period_data == "A" else period_data
    )

    return share_prices


def get_labour_productivity():
    """
    Labour productivity is defined as the ratio of a volume measure of GDP
    to a volume measure of labour input. This ratio is particularly useful
    when comparing international productivity levels. Labour productivity
    is also known as just productivity.

    Returns:
       pd.DataFrame: A DataFrame containing the labour productivity for a
         variety of countries over time.
    """
    oecd_data_string = ".GDPHRWKD.TOT.IDX2015.A"

    labour_productivity = collect_oecd_data(oecd_data_string, "Y")

    return labour_productivity


def get_population(gender: str | None = None):
    """
    Population is defined as all nationals present in, or temporarily absent
    from a country, and aliens permanently settled in a country.

    The number is presented in millions of people.

    Args:
        gender (str): specify the population based on gender.
            Defaults to None which returns the total population.

    Returns:
       pd.DataFrame: A DataFrame containing the population for a
            variety of countries over time.
    """
    if gender is not None and gender not in ["men", "women"]:
        raise ValueError("Please choose either 'men' or 'women'.")
    gender_parameter = (
        "MEN" if gender == "men" else "WOMEN" if gender == "women" else "TOT"
    )
    oecd_data_string = f".POP.{gender_parameter}.MLN_PER.A"

    population = collect_oecd_data(oecd_data_string, "Y")

    return population


def get_young_population_percentage():
    """
    The youth population is defined as those people aged less than 15.
    The share of the dependent population is calculated as total
    elderly and youth population expressed as a ratio of the total population.

    The number is presented in a % of the total population.

    Returns:
       pd.DataFrame: A DataFrame containing the young
        population for a variety of countries over time.
    """
    oecd_data_string = ".YNGPOP.TOT.PC_POP.A"

    young_population_percentage = collect_oecd_data(oecd_data_string, "Y")

    # Divide by 100 to get percentage
    young_population_percentage = young_population_percentage / 100

    return young_population_percentage


def get_working_age_population():
    """
    The working age population is defined as those aged 15 to 64.
    This indicator measures the share of the working age
    population in total population.

    Returns:
       pd.DataFrame: A DataFrame containing the working age
        population for a variety of countries over time.
    """
    oecd_data_string = ".WKGPOP.TOT.PC_POP.A"

    working_age_population = collect_oecd_data(oecd_data_string, "Y")

    # Divide by 100 to get percentage
    working_age_population = working_age_population / 100

    return working_age_population


def get_elderly_population_percentage():
    """
    The elderly population is defined as people aged 65 and over.
    The share of the dependent population is calculated as total
    elderly and youth population expressed as a ratio of
    the total population.

    The number is presented in a % of the total population.

    Returns:
       pd.DataFrame: A DataFrame containing the elderly
        population for a variety of countries over time.
    """
    oecd_data_string = ".ELDLYPOP.TOT.PC_POP.A"

    elderly_population_percentage = collect_oecd_data(oecd_data_string, "Y")

    # Divide by 100 to get percentage
    elderly_population_percentage = elderly_population_percentage / 100

    return elderly_population_percentage


def get_fertility_rate():
    """
    The total fertility rate in a specific year is defined as the
    total number of children that would be born to each woman if
    she were to live to the end of her child-bearing years and
    give birth to children in alignment with the prevailing
    age-specific fertility rates.

    Returns:
       pd.DataFrame: A DataFrame containing the fertility
        rate for a variety of countries over time.
    """
    oecd_data_string = ".FERTILITY.TOT.CHD_WOMAN.A"

    fertility_rate = collect_oecd_data(oecd_data_string, "Y")

    # Divide by 100 to get percentage
    fertility_rate = fertility_rate / 100

    return fertility_rate


def get_old_age_dependency_ratio():
    """
    The old-age to working-age demographic ratio is defined as
    the number of individuals aged 65 and over per 100 people
    of working age defined as those at ages 20 to 64.

    Returns:
       pd.DataFrame: A DataFrame containing the old-age
        dependency ratio for a variety of countries over time
    """
    oecd_data_string = ".OLDAGEDEP.TOT.PC.A"

    old_age_dependency = collect_oecd_data(oecd_data_string, "Y")

    # Divide by 100 to get percentage
    old_age_dependency = old_age_dependency / 100

    return old_age_dependency


def get_income_inequality(parameter: str = "GINI"):
    """
    Income is defined as household disposable income in a particular year.
    It consists of earnings, self-employment and capital income and public
    cash transfers; income taxes and social security contributions paid by
    households are deducted.

    Args:
        parameter (str): The parameter of the data. Can be 'GINI', 'P50P10',
        'P90P10', 'P90P50', 'S80S20' or 'PALMA'.

    Returns:
       pd.DataFrame: A DataFrame containing the income inequality
        for a variety of countries over time
    """
    parameter = parameter.upper()

    if parameter not in ["GINI", "P50P10", "P90P10", "P90P50", "S80S20", "PALMA"]:
        raise ValueError(
            "Please choose one of the following parameters: 'GINI', 'P50P10', "
            "'P90P10', 'P90P50', 'S80S20', 'PALMA'."
        )

    oecd_data_string = (
        f".INCOMEINEQ.{parameter}.{'INEQ' if parameter == 'GINI' else 'RT'}.A"
    )

    income_inequality = collect_oecd_data(oecd_data_string, "Y")

    return income_inequality


def get_poverty_rate(parameter: str = "TOT"):
    """
    Income is defined as household disposable income in a particular year.
    It consists of earnings, self-employment and capital income and public
    cash transfers; income taxes and social security contributions paid by
    households are deducted.

    Args:
        parameter (str): The parameter of the data. Can be 'TOT', '0_17',
        '18_65' or '66MORE'.

    Returns:
       pd.DataFrame: A DataFrame containing the income inequality
        for a variety of countries over time
    """
    parameter = parameter.upper()

    if parameter not in ["TOT", "0_17", "18_65", "66MORE"]:
        raise ValueError(
            "Please choose one of the following parameters: 'TOT', '0_17', "
            "'18_65', '66MORE'."
        )

    oecd_data_string = f".POVERTY.{parameter}.RT.A"

    poverty_rate = collect_oecd_data(oecd_data_string, "Y")

    return poverty_rate


def get_narrow_money(period: str):
    """
    M1 includes currency i.e. banknotes and coins, plus overnight deposits.
    M1 is expressed as a seasonally adjusted index based on 2015=100.

    Args:
        period (str): The period of the data. Can be 'monthly', 'quarterly' or 'yearly'.

    Returns:
       pd.DataFrame: A DataFrame containing the narrow money for a variety
        of countries over time.
    """
    period = period.lower()

    if period not in ["monthly", "quarterly", "yearly"]:
        raise ValueError("Period must be one of 'monthly', 'quarterly' or 'yearly'")

    period_data = "M" if period == "monthly" else "Q" if period == "quarterly" else "A"

    oecd_data_string = f".M1.TOT.IDX2015.{period_data}"

    narrow_money = collect_oecd_data(
        oecd_data_string, "Y" if period_data == "A" else period_data
    )

    return narrow_money


def get_broad_money(period: str):
    """
    Broad money (M3) includes currency, deposits with an agreed maturity of up
    to two years, deposits redeemable at notice of up to three months and
    repurchase agreements, money market fund shares/units and debt securities
    up to two years. M3 is measured as a seasonally adjusted index based on 2015=100.

    Args:
        period (str): The period of the data. Can be 'monthly', 'quarterly' or 'yearly'.

    Returns:
       pd.DataFrame: A DataFrame containing the broad money for a variety
        of countries over time.
    """
    period = period.lower()

    if period not in ["monthly", "quarterly", "yearly"]:
        raise ValueError("Period must be one of 'monthly', 'quarterly' or 'yearly'")

    period_data = "M" if period == "monthly" else "Q" if period == "quarterly" else "A"

    oecd_data_string = f".M3.TOT.IDX2015.{period_data}"

    broad_money = collect_oecd_data(
        oecd_data_string, "Y" if period_data == "A" else period_data
    )

    return broad_money


def get_general_government_deficit():
    """
    General government deficit is defined as the balance of income
    and expenditure of government, including capital income and
    capital expenditures.

    These metrics are expressed as a percentage of GDP.

    Returns:
       pd.DataFrame: A DataFrame containing the general government
        deficit for a variety of countries over time
    """
    oecd_data_string = ".GGNLEND.TOT.PC_GDP.A"

    general_government_deficit = collect_oecd_data(oecd_data_string, "Y")

    # Divide by 100 to get percentage
    general_government_deficit = general_government_deficit / 100

    return general_government_deficit


def get_general_government_revenue():
    """
    Governments collect revenues mainly for two purposes: to finance the
    goods and services they provide to citizens and businesses, and to
    fulfil their redistributive role.

    These metrics are expressed as a percentage of GDP.

    Returns:
       pd.DataFrame: A DataFrame containing the general government
        revenue for a variety of countries over time
    """
    oecd_data_string = ".GGREV.TOT.PC_GDP.A"

    general_government_revenue = collect_oecd_data(oecd_data_string, "Y")

    # Divide by 100 to get percentage
    general_government_revenue = general_government_revenue / 100

    return general_government_revenue


def get_general_government_spending(parameter: str = "TOT"):
    """
    General government spending provides an indication of the size of
    government across countries.

    General government final consumption can be broken down into two
    distinct groups. The first reflects expenditures for collective
    consumption (defence, justice, etc.) that benefit the society as
    a whole, or large parts of society, and are often known as
    public goods and services.

    These metrics are expressed as a percentage of GDP.

    Args:
        parameter (str): The parameter of the data. Can be 'TOT',
        'SOCIETY' or 'INDIVIDUAL'.

    Returns:
       pd.DataFrame: A DataFrame containing the general government
        spending for a variety of countries over time
    """
    parameter = parameter.upper()

    if parameter not in ["TOT", "SOCIETY", "INDIVIDUAL"]:
        raise ValueError(
            "Please choose one of the following parameters: 'TOT', 'SOCIETY', 'INDIVIDUAL'."
        )

    oecd_data_string = f".GGEXP.{parameter}.PC_GDP.A"

    general_government_spending = collect_oecd_data(oecd_data_string, "Y")

    # Divide by 100 to get percentage
    general_government_spending = general_government_spending / 100

    return general_government_spending


def get_general_government_debt():
    """
    General government debt-to-GDP ratio measures the
    gross debt of the general government as a percentage of GDP.

    These metrics are expressed as a percentage of GDP.

    Returns:
       pd.DataFrame: A DataFrame containing the general government
        debt for a variety of countries over time
    """
    oecd_data_string = ".GGDEBT.TOT.PC_GDP.A"

    general_government_debt = collect_oecd_data(oecd_data_string, "Y")

    # Divide by 100 to get percentage
    general_government_debt = general_government_debt / 100

    return general_government_debt


def get_general_government_financial_wealth():
    """
    The net financial worth of the general government sector is
    he total value of its financial assets minus the total value
    of its outstanding liabilities.

    These metrics are expressed as a percentage of GDP.

    Returns:
       pd.DataFrame: A DataFrame containing the general government
        financial wealth for a variety of countries over time
    """
    oecd_data_string = ".GGWEALTH.TOT.PC_GDP.A"

    general_government_financial_wealth = collect_oecd_data(oecd_data_string, "Y")

    # Divide by 100 to get percentage
    general_government_financial_wealth = general_government_financial_wealth / 100

    return general_government_financial_wealth


def get_general_government_production_costs():
    """
    General government production costs are decisions about
    the amount and type of goods and services governments produce,
    as well as on how best to produce them.

    These metrics are expressed as a percentage of GDP.

    Returns:
       pd.DataFrame: A DataFrame containing the general government
        financial wealth for a variety of countries over time
    """
    oecd_data_string = ".GGCOST.TOT.PC_GDP.A"

    general_government_production_costs = collect_oecd_data(oecd_data_string, "Y")

    # Divide by 100 to get percentage
    general_government_production_costs = general_government_production_costs / 100

    return general_government_production_costs


def get_central_government_spending(parameter: str):
    """
    Central government expenditure is defined as the central
    government budget expenditure as reported in the final
    central government accounts.

    This indicator of central government spending by function is
    measured as a percentage of total expenditures.

    Args:
        parameter (str): The parameter of the data. Can be 'GRALPUBSER', 'DEF', 'PUBORD',
            'ECOAFF', 'ENVPROT', 'HOUCOMM', 'HEALTH', 'RECULTREL', 'EDU', 'SOCPROT'.

    Returns:
       pd.DataFrame: A DataFrame containing the central government
        spending for a variety of countries over time
    """
    if parameter not in [
        "GRALPUBSER",
        "DEF",
        "PUBORD",
        "ECOAFF",
        "ENVPROT",
        "HOUCOMM",
        "HEALTH",
        "RECULTREL",
        "EDU",
        "SOCPROT",
    ]:
        raise ValueError(
            "Please choose one of the following parameters: 'GRALPUBSER', 'DEF', 'PUBORD', "
            "'ECOAFF', 'ENVPROT', 'HOUCOMM', 'HEALTH', 'RECULTREL', 'EDU', 'SOCPROT'."
        )

    oecd_data_string = f".CGEXP.{parameter}.PC.A"

    central_government_spending = collect_oecd_data(oecd_data_string, "Y")

    # Divide by 100 to get percentage
    central_government_spending = central_government_spending / 100

    return central_government_spending


def get_trust_in_goverment():
    """
    Trust in government refers to the share of people who report
    having confidence in the national government.

    Returns:
       pd.DataFrame: A DataFrame containing the trust in government
        for a variety of countries over time
    """
    oecd_data_string = ".TRUSTGOV.TOT.PC.A"

    trust_in_government = collect_oecd_data(oecd_data_string, "Y")

    # Divide by 100 to get percentage
    trust_in_government = trust_in_government / 100

    return trust_in_government


def get_renewable_energy():
    """
    Renewable energy is defined as the contribution of renewables
    to total primary energy supply (TPES).

    Returns:
       pd.DataFrame: A DataFrame containing the renewable energy
        for a variety of countries over time
    """
    oecd_data_string = ".RENEWABLE.TOT.PC_PRYENRGSUPPLY.A"

    renewable_energy = collect_oecd_data(oecd_data_string, "Y")

    # Divide by 100 to get percentage
    renewable_energy = renewable_energy / 100

    return renewable_energy


def get_crude_oil_production():
    """
    Crude oil production is defined as the quantities of oil extracted
    from the ground after the removal of inert matter or impurities.

    The number is presented in thousand tonnes of oil equivalent.

    Returns:
       pd.DataFrame: A DataFrame containing the crude oil production
        for a variety of countries over time
    """
    oecd_data_string = ".OILPROD.TOT.KTOE.A"

    crude_oil_production = collect_oecd_data(oecd_data_string, "Y")

    return crude_oil_production


def get_crude_oil_prices():
    """
    Crude oil import prices come from the IEA's Crude Oil Import Register
    and are influenced not only by traditional movements of supply and
    demand, but also by other factors such as geopolitics.

    The number is presented in US dollars per barrel.

    Returns:
       pd.DataFrame: A DataFrame containing the crude oil prices
        for a variety of countries over time
    """
    oecd_data_string = ".OILIMPPRICE.TOT.USD_BAR.A"

    crude_oil_prices = collect_oecd_data(oecd_data_string, "Y")

    return crude_oil_prices


def get_environmental_tax(parameter: str = "TOT"):
    """
    Environmentally related taxes are an important instrument for
    governments to shape relative prices of goods and services.

    Returns:
       pd.DataFrame: A DataFrame containing the environmental tax
        for a variety of countries over time
    """
    if parameter not in ["TOT", "ENRG", "TRANSPORT", "RES", "POL"]:
        raise ValueError(
            "Please choose one of the following parameters: 'TOT', 'ENRG', 'TRANSPORT', 'RES', 'POL'."
        )
    oecd_data_string = f".TAXENV.{parameter}.PC_GDP.A"

    environmental_tax = collect_oecd_data(oecd_data_string, "Y")

    # Divide by 100 to get percentage
    environmental_tax = environmental_tax / 100

    return environmental_tax


def get_greenhouse_emissions(parameter: str = "CO2"):
    """
    Greenhouse gases refer to the sum of seven gases that have direct effects on climate change:
    carbon dioxide (CO2), methane (CH4), nitrous oxide (N2O), chlorofluorocarbons (CFCs),
    hydrofluorocarbons (HFCs), perfluorocarbons (PFCs), sulphur hexafluoride (SF6) and
    nitrogen trifluoride (NF3).

    Returns:
       pd.DataFrame: A DataFrame containing the greenhouse emissions
        for a variety of countries over time
    """
    if parameter not in ["CO2", "CO", "GHG", "NOX", "SOX", "VOC"]:
        raise ValueError(
            "Please choose one of the following parameters: 'CO2', 'CO', 'GHG', 'NOX', 'SOX', 'VOC'."
        )

    kilo_or_tonnes = (
        "KG_CAP" if parameter in ["SOX", "CO", "NOX", "SOX", "VOC"] else "TONNE_CAP"
    )

    oecd_data_string = f".AIREMISSION.{parameter}.{kilo_or_tonnes}.A"

    greenhouse_emission = collect_oecd_data(oecd_data_string, "Y")

    # Divide by 1000 to get tonnes
    if kilo_or_tonnes == "KG_CAP":
        greenhouse_emission = greenhouse_emission / 1000

    return greenhouse_emission
